import {boundsToRect, getFocusRectangles, Rectangle} from './focusRectangles';

interface TrackMember {
  element: HTMLElement;
  bounds: Rectangle;
  closestDelta: number;
}

function centroid(bounds: Rectangle): [number, number] {
  return [bounds.point[0] + bounds.size[0] / 2, bounds.point[1] + bounds.size[1] / 2];
}

export function arrowFocus(baseElement: HTMLElement): () => void {
  let trackIndex: number;
  let activeCoord: number | undefined;
  let track: TrackMember[];

  function clickListener(event: MouseEvent) {
    if (event.target && 'tabIndex' in event.target) {
      activeCoord = undefined;
    }
  }

  function keydownListener(event: KeyboardEvent) {
    const activeElement = document.activeElement;
    if (activeElement === null) {
      return;
    }
    const activeBounds = boundsToRect(activeElement.getBoundingClientRect());
    const activeCentroid = centroid(activeBounds);

    let main: number;
    let delta: number;

    switch (event.key) {
      case 'Tab':
        activeCoord = undefined;
        return;
      case 'ArrowLeft':
        main = 0;
        delta = -1;
        break;
      case 'ArrowRight':
        main = 0;
        delta = 1;
        break;
      case 'ArrowUp':
        main = 1;
        delta = -1;
        break;
      case 'ArrowDown':
        main = 1;
        delta = 1;
        break;
      default:
        return;
    }

    if (main !== activeCoord) {
      activeCoord = main;
      const cross = 1 - main;
      const newTrack: TrackMember[] = [];
      getFocusRectangles(baseElement, (element, bounds) => {
        if (bounds.point[cross] + bounds.size[cross] <= activeBounds.point[cross]) {
          return;
        }
        if (bounds.point[cross] >= activeBounds.point[cross] + activeBounds.size[cross]) {
          return;
        }

        const closestDelta =
            Math.max(activeCentroid[cross] - (bounds.point[cross] + bounds.size[cross]),
                bounds.point[cross] - activeCentroid[cross]);
        newTrack.push({element, bounds, closestDelta});
      });

      newTrack.sort((a, b) => a.closestDelta - b.closestDelta);

      track = [];
      newTrack.forEach(prospect => {
        let index = 0;
        while (true) {
          if (index === track.length) {
            track.push(prospect);
            break;
          }
          const existing = track[index];
          if (existing.bounds.point[main] >=
              prospect.bounds.point[main] + prospect.bounds.size[main]) {
            track.splice(index, 0, prospect);
            break;
          }

          if (existing.bounds.point[main] + existing.bounds.size[main] >
              prospect.bounds.point[main]) {
            break;
          }
          index++;
        }
      });

      trackIndex = track.findIndex(trackMember => trackMember.element === activeElement);
    }

    const newIndex = trackIndex + delta;
    if (newIndex >= 0 && newIndex < track.length) {
      trackIndex = newIndex;
      track[trackIndex].element.focus();
      event.preventDefault();
    }
  }

  baseElement.addEventListener('click', clickListener);
  baseElement.addEventListener('keydown', keydownListener);

  return () => {
    baseElement.removeEventListener('click', clickListener);
    baseElement.removeEventListener('keydown', keydownListener);
  };

}
